/*
 * Copyright (C) Anthony Wright 2006
 *
 * Licensed according to the GPL (see COPYING)
 */
#include <ctype.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <linux/if_packet.h>

#include "dat.h"

#define DEFAULT_OPERATION		OperationRead
#define DEFAULT_PACKET_SECTOR_COUNT	2
#define DEFAULT_WINDOW_SIZE		25
#define DEFAULT_PACKET_COUNT		100000

#define TIME_DIFFERENCE(earlier, later) (later.tv_sec - earlier.tv_sec) + ((later.tv_usec - earlier.tv_usec) / 1000000.0)

#define SECTOR_SIZE			512

#define MAX_SHELF			0xfffe
#define MAX_SLOT			0xfe
#define MAX_PACKET_SECTOR_COUNT		(ATA_MAX_DATA / SECTOR_SIZE)
#define MAX_WINDOW_SIZE			10000
#define MAX_PACKET_COUNT		100000000000LL

#define ETHERTYPE_AOE			0x88a2

#define ATA_CMD_PIO_READ_EXT		0x24
#define ATA_CMD_PIO_WRITE_EXT		0x34
#define ATA_CMD_ID_ATA			0xEC

#define AOE_ATA_ID_SERIAL_OFFSET	20
#define AOE_ATA_ID_VERSION_OFFSET	46
#define AOE_ATA_ID_NAME_OFFSET		54
#define AOE_ATA_ID_LBA48_OFFSET		200

#define AOE_ATA_ID_SERIAL_SIZE		20
#define AOE_ATA_ID_VERSION_SIZE		8
#define AOE_ATA_ID_NAME_SIZE		40

#define LBA48_SIZE			6
#define AOE_TAG_SIZE			4
#define ETHERNET_MIN_PACKET_SIZE	60

enum Operation
{
	OperationRead,
	OperationWrite
} ;

struct Parameters
{
	char			*interface ;
	enum Operation		operation ;
	unsigned char		source[ETHER_ADDR_LEN] ;
	unsigned char		target[ETHER_ADDR_LEN] ;
	unsigned short		shelf ;
	unsigned char		slot ;
	unsigned char		packet_sector_count ;
	unsigned long		window_size ;
	unsigned long long	packet_count ;
	unsigned long long	disk_sectors ;

	unsigned short		packet_size ;

	unsigned long		tx_tag ;
} ;

struct ArgChoice
{
	char			*string ;
	char			character ;
	int			value ;
} ;

static void			Usage(void) ;

static void			ParseArgs(struct Parameters *, int, char **) ;
static int			ParseArgChoice(char *, char *, struct ArgChoice *) ;
static long long		ParseArgInteger(char *, char *, long long, long long) ;

static int			OpenNetwork(struct Parameters *) ;
static void			FindServer(int, struct Parameters *) ;
static void			Transfer(int, struct Parameters *) ;

static void			EncodeTag(unsigned char *, unsigned long) ;
static unsigned long		DecodeTag(unsigned char *) ;
static void			EncodeLBA48(unsigned char *, unsigned long long) ;
static unsigned long long	DecodeLBA48(unsigned char *) ;
static void			EncodeValue(unsigned char *, int, unsigned long long) ;
static unsigned long long	DecodeValue(unsigned char *, int) ;

static void			DecodeString(unsigned char *, int, char *) ;

static struct ArgChoice	ArgChoiceOperations[] =
{
	{ "read",	'r', OperationRead	},
	{ "write",	'w', OperationWrite	},
	{ 0 }
} ;

static unsigned char	broadcast_address[ETHER_ADDR_LEN] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } ;

int main(int argc, char **argv)
{
	struct Parameters	parameters ;
	int			sock ;

	parameters.tx_tag = 0 ;

	ParseArgs(&parameters, argc, argv) ;

	sock = OpenNetwork(&parameters) ;

	FindServer(sock, &parameters) ;

	Transfer(sock, &parameters) ;

	close(sock) ;

	return(0) ;
}

static void Usage()
{
	fprintf(stderr, "usage: disktest [options] <interface> <shelf> <slot>\n") ;
	fprintf(stderr, "\n") ;
	fprintf(stderr, "    -o --operation    read|write r|w (Default: %s)\n",   ArgChoiceOperations[DEFAULT_OPERATION].string) ;
	fprintf(stderr, "    -s --sector-count <sectors>      (Default: %llu)\n", (unsigned long long)DEFAULT_PACKET_SECTOR_COUNT) ;
	fprintf(stderr, "    -j --jumbo                       (equivalent to --sector-count 17)\n") ;
	fprintf(stderr, "    -w --window-size  <packets>      (Default: %lu)\n",  (unsigned long)DEFAULT_WINDOW_SIZE) ;
	fprintf(stderr, "    -p --packet-count <packets>      (Default: %llu)\n", (unsigned long long)DEFAULT_PACKET_COUNT) ;
	exit(1) ;
}

static void ParseArgs(struct Parameters *parameters, int argc, char **argv)
{
	char	option_string[1024] ;
	char	option_char ;
	char	*value ;

	parameters->operation           = DEFAULT_OPERATION ;
	parameters->packet_sector_count = DEFAULT_PACKET_SECTOR_COUNT ;
	parameters->window_size         = DEFAULT_WINDOW_SIZE ;
	parameters->packet_count        = DEFAULT_PACKET_COUNT ;

	for(argc--, argv++; argc > 0; argc--, argv++)
	{
		strcpy(option_string, *argv) ;

		if (option_string[0] != '-')
			break ;

		option_char = (option_string[1] != '-') ? option_string[1] : 0 ;

		if ((option_char == 'j') || !strcmp(option_string, "--jumbo"))
		{
			parameters->packet_sector_count = 17 ;
			continue ;
		}

		if ((value = strchr(option_string, '=')))
		{
			*(value++) = '\0' ;
		}
		else
		{
			if (argc == 1)
				Usage() ;

			argc-- ;
			argv++ ;

			value = *argv ;
		}

		if ((option_char == 'o') || !strcmp(option_string, "--operation"))
		{
			parameters->operation = ParseArgChoice(value, "operation", ArgChoiceOperations) ;
			continue ;
		}

		if ((option_char == 's') || !strcmp(option_string, "--sector-count"))
		{
			parameters->packet_sector_count = (unsigned long long)ParseArgInteger(value, "sector count", 0, MAX_PACKET_SECTOR_COUNT) ;
			continue ;
		}

		if ((option_char == 'w') || !strcmp(option_string, "--window-size"))
		{
			parameters->window_size = (unsigned long)ParseArgInteger(value, "window size", 0, MAX_WINDOW_SIZE) ;
			continue ;
		}

		if ((option_char == 'p') || !strcmp(option_string, "--packet-count"))
		{
			parameters->packet_count = (unsigned long long)ParseArgInteger(value, "packet count", 0, MAX_PACKET_COUNT) ;
			continue ;
		}

		Usage() ;
	}

	if (argc != 3)
		Usage() ;

	parameters->interface = argv[0] ;

	parameters->shelf = (unsigned short)ParseArgInteger(argv[1], "shelf", 0, MAX_SHELF) ;
	parameters->slot  = (unsigned char) ParseArgInteger(argv[2], "slot",  0, MAX_SLOT) ;

	printf("Interface:          %s\n",   parameters->interface) ;
	printf("Shelf:              %u\n",   parameters->shelf) ;
	printf("Slot:               %u\n",   parameters->slot) ;
	printf("Operation:          %s\n",   ArgChoiceOperations[parameters->operation].string) ;
	printf("Sector count:       %u\n",   parameters->packet_sector_count) ;
	printf("Window size:        %lu\n",  parameters->window_size) ;
	printf("Packet count:       %llu\n", parameters->packet_count) ;

	parameters->packet_size = sizeof(Ata) - ATA_MAX_DATA + (parameters->packet_sector_count * SECTOR_SIZE) ;
}

static int ParseArgChoice(char *arg, char *name, struct ArgChoice *choices)
{
	struct ArgChoice	*choice ;

	for(choice = choices; choice->string; choice++)
	{
		if (!strcmp(choice->string, arg))
			return(choice->value) ;

		if ((strlen(arg) == 1) && (choice->character == arg[0]))
			return(choice->value) ;
	}

	fprintf(stderr, "Unknown %s: %s\n", name, arg) ;
	exit(1) ;
}

static long long ParseArgInteger(char *arg, char *name, long long min, long long max)
{
	long long value ;

	if ((value = atoll(arg)) == 0)
	{
		fprintf(stderr, "Invalid %s: %s\n", name, arg) ;
		exit(1) ;
	}

	if ((value < min) || (value > max))
	{
		fprintf(stderr, "Invalid %s value: %lld (out of range)\n", name, value) ;
		exit(1) ;
	}

	return(value) ;
}

static int OpenNetwork(struct Parameters *parameters)
{
	struct sockaddr_ll	sockaddr ;
	struct ifreq		ifreq ;
	int			sock ;

	memset(&sockaddr, 0, sizeof(sockaddr)) ;

	if ((sock = socket(PF_PACKET, SOCK_RAW, htons(ETHERTYPE_AOE))) == -1)
	{
		perror("socket") ;
		exit(1) ;
	}

	strcpy(ifreq.ifr_name, parameters->interface) ;

	if (ioctl(sock, SIOCGIFHWADDR, &ifreq) == -1)
	{
		perror("ioctl(SIOCGIFHWADDR)") ;
		exit(1) ;
	}

	memcpy(parameters->source, ifreq.ifr_hwaddr.sa_data, IFHWADDRLEN) ;

	if (ioctl(sock, SIOCGIFINDEX, &ifreq) == -1)
	{
		perror("ioctl(SIOCGIFINDEX)") ;
		exit(1) ;
	}

	sockaddr.sll_family   = AF_PACKET ;
	sockaddr.sll_protocol = htons(ETHERTYPE_AOE) ;
	sockaddr.sll_ifindex  = ifreq.ifr_ifindex ;

	if (bind(sock, (struct sockaddr *)&sockaddr, sizeof(sockaddr)) == -1)
	{
		perror("bind") ;
		exit(1) ;
	}

	if (ioctl(sock, SIOCGIFMTU, &ifreq) == -1)
	{
		perror("ioctl(SIOCGIFHWADDR)") ;
		exit(1) ;
	}

	printf("\n") ;
	printf("Packet Size:        %u Bytes\n", parameters->packet_size) ;
	printf("Interface MTU:      %u Bytes\n", ifreq.ifr_ifru.ifru_mtu) ;

	if (parameters->packet_size > ifreq.ifr_ifru.ifru_mtu)
	{
		fprintf(stderr, "Error: Packet is too large to be transmitted on the interface\n") ;
		exit(1) ;
	}

	return(sock) ;
}

static void FindServer(int sock, struct Parameters *parameters)
{
	Ata			tx_packet ;
	Ata			rx_packet ;
	struct timeval		start_time ;
	struct timeval		now ;
	double			elapsed ;
	struct pollfd		poll_fds[1] ;
	char			name[AOE_ATA_ID_NAME_SIZE + 1] ;
	char			version[AOE_ATA_ID_VERSION_SIZE + 1] ;
	char			serial[AOE_ATA_ID_SERIAL_SIZE + 1] ;

	memcpy(tx_packet.h.src, parameters->source, sizeof(tx_packet.h.src)) ;
	memcpy(tx_packet.h.dst, broadcast_address,  sizeof(tx_packet.h.dst)) ;
	tx_packet.h.type   = htons(ETHERTYPE_AOE) ;
	tx_packet.h.flags  = 0x10 ; // Version:4 + Flags:4
	tx_packet.h.error  = 0 ;
	tx_packet.h.maj    = htons(parameters->shelf) ;
	tx_packet.h.min    = parameters->slot ;
	tx_packet.h.cmd    = ATAcmd ;

	tx_packet.aflag    = 0 ;
	tx_packet.err      = 0 ;
	tx_packet.sectors  = 1 ;
	tx_packet.cmd      = ATA_CMD_ID_ATA ;

	poll_fds[0].fd     = sock ;
	poll_fds[0].events = POLLIN ;

	for(;;)
	{
		EncodeTag(tx_packet.h.tag, parameters->tx_tag++) ;

		if (write(sock, &tx_packet, ETHERNET_MIN_PACKET_SIZE) == -1)
		{
			perror("FindServer write") ;
			exit(1) ;
		}

		gettimeofday(&start_time, 0) ;

		for(;;)
		{
			gettimeofday(&now, 0) ;

			elapsed = TIME_DIFFERENCE(start_time, now) ;

			if (elapsed >= 1)
			{
				printf("Unable to contact AOE server, retrying...\n") ;
				break ;
			}

			if (poll(poll_fds, 1, (int)((1 - elapsed) * 1000) + 10) == 0)
				continue ;

			if (read(sock, &rx_packet, parameters->packet_size) == -1)
			{
				perror("FindServer read") ;
				exit(1) ;
			}

			if (rx_packet.h.type != htons(ETHERTYPE_AOE))
				continue ;

			if (!(rx_packet.h.flags & Resp))
				continue ;

			if (rx_packet.h.maj != htons(parameters->shelf))
				continue ;

			if (rx_packet.h.min != parameters->slot)
				continue ;

			if (rx_packet.h.cmd != ATAcmd)
				continue ;

			if (DecodeTag(rx_packet.h.tag) != (parameters->tx_tag - 1))
				continue ;

			break ;
		}

		if (poll_fds[0].revents & POLLIN)
			break ;
	}

	memcpy(parameters->target, rx_packet.h.src, sizeof(rx_packet.h.src)) ;

	parameters->disk_sectors = DecodeLBA48(&rx_packet.data[AOE_ATA_ID_LBA48_OFFSET]) ;

	if ((parameters->disk_sectors % parameters->packet_sector_count) != 0)
	{
		printf("WARNING: Disk sectors not multiple of %d, will not access last sectors.\n", parameters->packet_sector_count) ;
		parameters->disk_sectors -= parameters->disk_sectors % parameters->packet_sector_count ;
	}

	DecodeString(&rx_packet.data[AOE_ATA_ID_NAME_OFFSET],    AOE_ATA_ID_NAME_SIZE,    name) ;
	DecodeString(&rx_packet.data[AOE_ATA_ID_VERSION_OFFSET], AOE_ATA_ID_VERSION_SIZE, version) ;
	DecodeString(&rx_packet.data[AOE_ATA_ID_SERIAL_OFFSET],  AOE_ATA_ID_SERIAL_SIZE,  serial) ;

	printf("\n") ;
	printf("Server address:     %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n", parameters->target[0], parameters->target[1], parameters->target[2], parameters->target[3], parameters->target[4], parameters->target[5]) ;
	printf("Disk name:          %s\n", name) ;
	printf("Disk version:       %s\n", version) ;
	printf("Disk serial number: %s\n", serial) ;
	printf("Disk size:          %.2f MB (%.2f GB)\n", (double)parameters->disk_sectors / 2 / 1024, (double)parameters->disk_sectors / 2 / 1024 / 1024) ;
	printf("\n") ;
}

static void Transfer(int sock, struct Parameters *parameters)
{
	Ata			tx_packet ;
	Ata			rx_packet ;
	struct timeval		start_time ;
	struct timeval		last_time ;
	struct timeval		now ;
	unsigned long long	last_tx_count ;
	unsigned long long	last_rx_count ;
	unsigned long long	last_sector_count ;
	unsigned long long	last_data_errors ;
	unsigned long long	tx_count ;
	unsigned long long	rx_count ;
	unsigned long long	sector_count ;
	unsigned long long	data_errors ;
	unsigned long long	tx_sector ;
	unsigned long long	rx_sector ;
	struct pollfd		poll_fds[1] ;
	double			elapsed ;
	double			interval ;
	int			offset ;

	tx_sector = 0 ;
	tx_count = 0 ;
	rx_count = 0 ;
	sector_count = 0 ;
	data_errors = 0 ;

	memcpy(tx_packet.h.src, parameters->source, sizeof(tx_packet.h.src)) ;
	memcpy(tx_packet.h.dst, parameters->target, sizeof(tx_packet.h.dst)) ;
	tx_packet.h.type   = htons(ETHERTYPE_AOE) ;
	tx_packet.h.flags  = 0x10 ; // Version:4 + Flags:4
	tx_packet.h.error  = 0 ;
	tx_packet.h.maj    = htons(parameters->shelf) ;
	tx_packet.h.min    = parameters->slot ;
	tx_packet.h.cmd    = ATAcmd ;

	switch(parameters->operation)
	{
	case OperationRead:
		tx_packet.aflag    = Extend ;
		tx_packet.err      = 0 ;
		tx_packet.sectors  = parameters->packet_sector_count ;
		tx_packet.cmd      = ATA_CMD_PIO_READ_EXT ;
		break ;

	case OperationWrite:
		tx_packet.aflag    = Extend | Write ;
		tx_packet.err      = 0 ;
		tx_packet.sectors  = parameters->packet_sector_count ;
		tx_packet.cmd      = ATA_CMD_PIO_WRITE_EXT ;

		memset(tx_packet.data, 0, sizeof(tx_packet.data)) ;
		break ;
	}

	poll_fds[0].fd     = sock ;
	poll_fds[0].events = POLLIN ;

	last_tx_count     = 0 ;
	last_rx_count     = 0 ;
	last_sector_count = 0 ;
	last_data_errors  = 0 ;

	printf("Elapsed Tx-Pkt/s Rx-Pkt/s   MB/s");

	switch(parameters->operation)
	{
	case OperationRead:
		printf(" Data-Error/s") ;
		break ;

	default:
		break ;
	}

	printf("\n") ;

	gettimeofday(&start_time, 0) ;

	last_time = start_time ;

	for(;;)
	{
		gettimeofday(&now, 0) ;

		interval = TIME_DIFFERENCE(last_time, now) ;

		if (interval >= 1)
		{
			elapsed = TIME_DIFFERENCE(start_time, now) ;

			printf("%7.2f %8.f %8.f %6.2f",
				elapsed,
				(double)(tx_count - last_tx_count) / interval,
				(double)(rx_count - last_rx_count) / interval,
				(double)(sector_count - last_sector_count) / 2 / 1024 / interval) ;

			switch(parameters->operation)
			{
			case OperationRead:
				printf(" %12.0f", (double)(data_errors - last_data_errors) / interval) ;
				break ;

			default:
				break ;
			}

			printf("\n") ;

			last_time         = now ;
			last_tx_count     = tx_count ;
			last_rx_count     = rx_count ;
			last_sector_count = sector_count ;
			last_data_errors  = data_errors ;
		}

		if (poll(poll_fds, 1, 0) == 0)
		{
			if (tx_count < parameters->packet_count)
			{
				if ((tx_count - rx_count) < parameters->window_size)
				{
					EncodeLBA48(tx_packet.lba, tx_sector) ;

					EncodeTag(tx_packet.h.tag, parameters->tx_tag++) ;

					switch(parameters->operation)
					{
					case OperationRead:
						if (write(sock, &tx_packet, ETHERNET_MIN_PACKET_SIZE) == -1)
						{
							perror("write") ;
							exit(1) ;
						}

						break ;

					case OperationWrite:
						for(offset = 0 ; offset < parameters->packet_sector_count; offset++)
						{
							EncodeLBA48(&tx_packet.data[  offset      * SECTOR_SIZE],               tx_sector + offset) ;
							EncodeLBA48(&tx_packet.data[((offset + 1) * SECTOR_SIZE) - LBA48_SIZE], tx_sector + offset) ;
						}

						if (write(sock, &tx_packet, parameters->packet_size) == -1)
						{
							perror("Transfer write") ;
							exit(1) ;
						}

						break ;
					}

					tx_count++ ;
					tx_sector += parameters->packet_sector_count ;

					if (tx_sector == parameters->disk_sectors)
						tx_sector = 0 ;

					continue ;
				}
			}

			gettimeofday(&now, 0) ;

			if (poll(poll_fds, 1, 1000) == 0)
				break ;
		}

		if (read(sock, &rx_packet, parameters->packet_size) == -1)
		{
			perror("Transfer read") ;
			exit(1) ;
		}

		if (rx_packet.h.type != htons(ETHERTYPE_AOE))
			continue ;

		if (!(rx_packet.h.flags & Resp))
			continue ;

		if (rx_packet.h.maj != htons(parameters->shelf))
			continue ;

		if (rx_packet.h.min != parameters->slot)
			continue ;

		if (rx_packet.h.cmd != ATAcmd)
			continue ;

		switch(parameters->operation)
		{
		case OperationRead:
			rx_sector = DecodeLBA48(rx_packet.lba) ;

			for(offset = 0 ; offset < parameters->packet_sector_count; offset++)
			{
				if (DecodeLBA48(&rx_packet.data[  offset      * SECTOR_SIZE])               != rx_sector + offset)
				{
					data_errors++ ;
					continue ;
				}

				if (DecodeLBA48(&rx_packet.data[((offset + 1) * SECTOR_SIZE) - LBA48_SIZE]) != rx_sector + offset)
				{
					data_errors++ ;
					continue ;
				}
			}

			break ;

		case OperationWrite:
			break ;
		}

		rx_count++ ;
		sector_count += parameters->packet_sector_count ;
	}

	elapsed = TIME_DIFFERENCE(start_time, now) ;

	printf("\n") ;
	printf("Elapsed time:         %.2f s\n",     elapsed) ;
	printf("Packets sent:         %llu\n",       tx_count) ;
	printf("Packets received:     %llu\n",       rx_count) ;
	printf("Data transferred:     %.2f MB\n",    (double)sector_count / 2 / 1024) ;
	printf("Packet transmit rate: %.2f Pkt/s\n", tx_count / elapsed) ;
	printf("Packet receive rate:  %.2f Pkt/s\n", rx_count / elapsed) ;
	printf("Data transfer rate:   %.2f MB/s\n",  ((double)sector_count / 2 / 1024) / elapsed) ;

	switch(parameters->operation)
	{
	case OperationRead:
		printf("Data errors:          %llu\n",      data_errors) ;
		break ;

	default:
		break ;
	}
}

static void EncodeTag(unsigned char *buf, unsigned long tag)
{
	EncodeValue(buf, AOE_TAG_SIZE, tag) ;
}

static unsigned long DecodeTag(unsigned char *buf)
{
	return((unsigned long)DecodeValue(buf, AOE_TAG_SIZE)) ;
}

static void EncodeLBA48(unsigned char *buf, unsigned long long lba)
{
	EncodeValue(buf, LBA48_SIZE, lba) ;
}

static unsigned long long DecodeLBA48(unsigned char *buf)
{
	return(DecodeValue(buf, LBA48_SIZE)) ;
}

static void EncodeValue(unsigned char *buf, int size, unsigned long long value)
{
	int i ;

	for(i = 0; i < size; i++)
	{
		*(buf++) = value & 0xff ;
		value >>= 8 ;
	}
}

static unsigned long long DecodeValue(unsigned char *buf, int size)
{
	unsigned long long	value ;
	int			i ;

	value = 0 ;
	buf += (size - 1) ;

	for(i = 0; i < size; i++)
		value = (value << 8) | *(buf--) ;

	return(value) ;
}

static void DecodeString(unsigned char *buf, int size, char *string)
{
	char	*ptr ;

	ptr = string ;

	while(size > 0)
	{
		ptr[0] = buf[1] ;
		ptr[1] = buf[0] ;
		
		buf += 2 ;
		ptr += 2 ;

		size -=2 ;
	}

	*ptr = '\0' ;

	for(ptr = string; *ptr; ptr++)
		if (!isprint(*ptr))
			*ptr = '?' ;
}
