/*
 * Copyright (C) Anthony Wright 2006
 *
 * Licensed according to the GPL (see COPYING)
 */
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/mount.h>
#include <sys/shm.h>
#include <sys/time.h>

#define DEFAULT_OPERATION	OperationRead
#define DEFAULT_MODE		ModeDirect
#define DEFAULT_CHUNK_SIZE	20480
#define DEFAULT_DURATION	10

#ifndef O_DIRECT
#define O_DIRECT	040000	/* direct disk access, not easily obtained from headers */
#endif

#define TIME_DIFFERENCE(earlier, later) (later.tv_sec - earlier.tv_sec) + ((later.tv_usec - earlier.tv_usec) / 1000000.0)

#define MAX_RUN_SECONDS		1000000000

#define SECTOR_SIZE		512
#define LBA48_SIZE		6

enum Operation
{
	OperationRead = 0,
	OperationWrite,
	OperationVerifiedWrite
} ;

enum Mode
{
	ModeDirect = 0,
	ModeBuffered,
	ModeBufferedSync
} ;

struct Parameters
{
	enum Operation		operation ;
	enum Mode		mode ;
	size_t			chunk_size ;
	double			duration ;
	char			*device ;
	unsigned long long	device_size ;
} ;

struct ArgChoice
{
	char			*string ;
	char			character ;
	int			value ;
} ;

static void		Usage(void) ;

static void		ParseArgs(struct Parameters *, int, char **) ;
static int		ParseArgChoice(char *, char *, struct ArgChoice *) ;

static int		OpenDisk(struct Parameters *) ;
static void		TestDisk(int, struct Parameters *) ;

static void		Read(int, unsigned char *, size_t, unsigned long long, unsigned long long, unsigned long long *) ;
static void		Write(int, unsigned char *, size_t, unsigned long long, unsigned long long) ;

void			EncodeLBA48(unsigned char *, unsigned long long) ;
unsigned long long	DecodeLBA48(unsigned char *) ;
void			EncodeValue(unsigned char *, int, unsigned long long) ;
unsigned long long	DecodeValue(unsigned char *, int) ;

static struct ArgChoice	ArgChoiceOperations[] =
{
	{ "read",		'r', OperationRead		},
	{ "write",		'w', OperationWrite		},
	{ "verified-write",	'v', OperationVerifiedWrite	},
	{ 0 }
} ;

static struct ArgChoice	ArgChoiceModes[] =
{
	{ "direct",		'd', ModeDirect		},
	{ "buffered",		'b', ModeBuffered	},
	{ "buffered-sync",	's', ModeBufferedSync	},
	{ 0 }
} ;

int main(int argc, char **argv)
{
	struct Parameters	parameters ;
	int			fd ;

	ParseArgs(&parameters, argc, argv) ;

	fd = OpenDisk(&parameters) ;
	TestDisk(fd, &parameters) ;
	close(fd) ;

	return(0) ;
}

static void Usage()
{
	fprintf(stderr, "usage: disktest [options] <device>\n") ;
	fprintf(stderr, "\n") ;
	fprintf(stderr, "    -o --operation  read|write|verified-write     r|w|v (Default: %s)\n",   ArgChoiceOperations[DEFAULT_OPERATION].string) ;
	fprintf(stderr, "    -m --mode       direct|buffered|buffered-sync d|b|s (Default: %s)\n",   ArgChoiceModes[DEFAULT_MODE].string) ;
	fprintf(stderr, "    -c --chunk-size <size>                              (Default: %lu)\n",  (unsigned long)DEFAULT_CHUNK_SIZE) ;
	fprintf(stderr, "    -d --duration   <time>                              (Default: %.1f)\n", (double)DEFAULT_DURATION) ;
	exit(1) ;
}

void ParseArgs(struct Parameters *parameters, int argc, char **argv)
{
	char	option_string[1024] ;
	char	option_char ;
	char	*value ;

	parameters->operation  = DEFAULT_OPERATION ;
	parameters->mode       = DEFAULT_MODE ;
	parameters->chunk_size = DEFAULT_CHUNK_SIZE ;
	parameters->duration   = DEFAULT_DURATION ;


	for(argc--, argv++; argc > 0; argc--, argv++)
	{
		strcpy(option_string, *argv) ;

		if (option_string[0] != '-')
			break ;

		option_char = (option_string[1] != '-') ? option_string[1] : 0 ;

		if ((value = strchr(option_string, '=')))
		{
			*(value++) = '\0' ;
		}
		else
		{
			if (argc == 1)
				Usage() ;

			argc-- ;
			argv++ ;

			value = *argv ;
		}

		if ((option_char == 'o') || !strcmp(option_string, "--operation"))
		{
			parameters->operation = ParseArgChoice(value, "operation", ArgChoiceOperations) ;
			continue ;
		}

		if ((option_char == 'm') || !strcmp(option_string, "--mode"))
		{
			parameters->mode = ParseArgChoice(value, "mode", ArgChoiceModes) ;
			continue ;
		}

		if ((option_char == 'c') || !strcmp(option_string, "--chunk-size"))
		{
			if ((parameters->chunk_size = atol(value)) == 0)
			{
				fprintf(stderr, "Invalid chunk size: %s\n", value) ;
				exit(1) ;
			}

			if ((parameters->chunk_size % SECTOR_SIZE) != 0)
			{
				fprintf(stderr, "Chunk size (%u) not a multiple of %d\n", parameters->chunk_size, SECTOR_SIZE) ;
				exit(1) ;
			}

			continue ;
		}

		if ((option_char == 'd') || !strcmp(option_string, "--duration"))
		{
			if ((parameters->duration = atof(value)) == 0)
			{
				fprintf(stderr, "Invalid duration: %s\n", value) ;
				exit(1) ;
			}

			continue ;
		}

		Usage() ;
	}

	if (argc != 1)
		Usage() ;

	parameters->device = *argv ;
}

static int ParseArgChoice(char *arg, char *name, struct ArgChoice *choices)
{
	struct ArgChoice	*choice ;

	for(choice = choices; choice->string; choice++)
	{
		if (!strcmp(choice->string, arg))
			return(choice->value) ;

		if ((strlen(arg) == 1) && (choice->character == arg[0]))
			return(choice->value) ;
	}

	fprintf(stderr, "Unknown %s: %s\n", name, arg) ;
	exit(1) ;
}

static int OpenDisk(struct Parameters *parameters)
{
	int			open_flags ;
	unsigned long long	device_size64 ;
	unsigned long		device_size32 ;
	int			fd ;

	open_flags = 0 ;

	switch(parameters->operation)
	{
	case OperationRead:
		open_flags = O_RDONLY ;
		break ;

	case OperationWrite:
		open_flags = O_WRONLY ;
		break ;

	case OperationVerifiedWrite:
		open_flags = O_RDWR ;
		break ;
	}

	if (parameters->mode == ModeDirect)
		open_flags |= O_DIRECT ;

	if ((fd = open(parameters->device, open_flags)) < 0)
	{
		perror(parameters->device) ;
		exit(1) ;
	}

	if (ioctl(fd, BLKGETSIZE64, &device_size64) == 0)
	{
		parameters->device_size = device_size64 ;

		if ((parameters->device_size % SECTOR_SIZE) != 0)
		{
			printf("WARNING: Disk size not a multiple of %d.\n", SECTOR_SIZE) ;
			parameters->device_size -= (parameters->device_size % SECTOR_SIZE) ;
		}
	}
	else
	{
		if (ioctl(fd, BLKGETSIZE, &device_size32) != 0)
		{
			perror(" BLKGETSIZE failed") ;
			exit(1) ;
		}

		parameters->device_size = device_size32 * SECTOR_SIZE ;
	}

	if (parameters->chunk_size > parameters->device_size)
	{
		printf("WARNING: chunk-size is larger than disk size, reducing chunk-size\n") ;
		parameters->chunk_size = parameters->device_size ;
	}

	return(fd) ;
}

static void TestDisk(int fd, struct Parameters *parameters)
{
	unsigned char		*memory ;
	unsigned char		*buf ;
	unsigned long long	position ;
	unsigned long long	transfer_total ;
	unsigned long long	verify_code ;
	unsigned long long	data_errors ;
	unsigned long long	last_transfer_total ;
	unsigned long long	last_data_errors ;
	size_t			transfer_size ;
	struct timeval		start_time ;
	struct timeval		last_stats_time ;
	struct timeval		end_time ;
	struct timeval		sync_overhead_time ;
	struct timeval		now ;
	double			elapsed ;
	double			interval ;

	if (!(memory = malloc(parameters->chunk_size + SECTOR_SIZE - 1)))
	{
		fprintf(stderr, "could not not allocate %u bytes of memory\n", parameters->chunk_size) ;
		exit(1) ;
	}

	/* Aligning buf on SECTOR_SIZE boundary - required for direct I/O */
	buf = memory + SECTOR_SIZE - 1 ;
	buf = buf - (((unsigned long)buf) % SECTOR_SIZE) ;

	memset(buf, 0, parameters->chunk_size) ;

	printf("Operation:  %s\n",     ArgChoiceOperations[parameters->operation].string) ;
	printf("Mode:       %s\n",     ArgChoiceModes[parameters->mode].string) ;
	printf("Chunk size: %u\n",     parameters->chunk_size) ;
	printf("Duration:   %.2f s\n", parameters->duration) ;
	printf("Disk:       %s\n",     parameters->device) ;
	printf("Disk size:  %.2f MB (%.2f GB)\n", (double)parameters->device_size / 1024 / 1024, (double)parameters->device_size / 1024 / 1024 / 1024) ;
	printf("\n") ;

	sync() ;
	sleep(3) ;

	printf("Elapsed   MB/s") ;

	switch(parameters->operation)
	{
	case OperationRead:
	case OperationVerifiedWrite:
		printf(" Data-Error/s") ;
		break ;

	default:
		break ;
	}

	printf("\n") ;

	srand48(time(0)) ;

	position       = 0 ;
	transfer_total = 0 ;
	data_errors    = 0 ;

	gettimeofday(&start_time, 0) ;

	last_stats_time     = start_time ;
	last_transfer_total = 0 ;
	last_data_errors    = 0 ;

	do
	{
		if (parameters->chunk_size <= (parameters->device_size - position))
			transfer_size = parameters->chunk_size ;
		else
			transfer_size = parameters->device_size - position ;

		switch(parameters->operation)
		{
		case OperationRead:
			Read(fd, buf, transfer_size, position, 0, &data_errors) ;
			break ;

		case OperationWrite:
			Write(fd, buf, transfer_size, position, 0) ;
			break ;

		case OperationVerifiedWrite:
			verify_code = lrand48() ;

			Write(fd, buf, transfer_size, position, verify_code) ;
			lseek(fd, position, SEEK_SET) ;
			Read(fd, buf, transfer_size, position, verify_code, &data_errors) ;
			break ;
		}

		transfer_total += transfer_size ;

		position += transfer_size ;

		if (position == parameters->device_size)
		{
			lseek(fd, 0, SEEK_SET) ;
			position = 0 ;
		}

		gettimeofday(&now, 0) ;

		elapsed  = TIME_DIFFERENCE(start_time, now) ;
		interval = TIME_DIFFERENCE(last_stats_time, now) ;

		if (interval >= 1)
		{
			printf("%7.2f %6.2f",
				elapsed,
				(double)(transfer_total - last_transfer_total) / 1024 / 1024 / interval) ;

			switch(parameters->operation)
			{
			case OperationRead:
			case OperationVerifiedWrite:
				printf(" %12.0f", (double)(data_errors - last_data_errors) / interval) ;
				break ;

			default:
				break ;
			}

			printf("\n") ;

			last_stats_time     = now ;
			last_transfer_total = transfer_total ;
			last_data_errors    = data_errors ;
		}
	}
	while (elapsed < parameters->duration) ;

	if (parameters->mode == ModeBufferedSync)
	{
		sync() ;

		gettimeofday(&end_time, 0) ;

		sync() ;

		gettimeofday(&sync_overhead_time, 0) ;

		elapsed = TIME_DIFFERENCE(start_time, end_time) - TIME_DIFFERENCE(end_time, sync_overhead_time) ;
	}

	printf("\n") ;
	printf("Elapsed time:       %.2f s\n",    elapsed) ;
	printf("Data transferred:   %.2f MB\n",   (double)transfer_total / 1024 / 1024) ;
	printf("Data transfer rate: %.2f MB/s\n", (double)transfer_total / 1024 / 1024 / elapsed) ;

	switch(parameters->operation)
	{
	case OperationRead:
	case OperationVerifiedWrite:
		printf("Data errors:        %llu\n",      data_errors) ;
		break ;

	default:
		break ;
	}

	free(memory) ;

	close(fd) ;
}

static void Read(int fd, unsigned char *buf, size_t transfer_size, unsigned long long position, unsigned long long verify_code, unsigned long long *data_errors)
{
	unsigned long long	sector ;
	size_t			offset ;
	ssize_t			ret ;

	if ((ret = read(fd, buf, transfer_size)) != transfer_size)
	{
		if (ret < 0)
			perror("read() failed") ;
		else
			fprintf(stderr, "read(%u) returned %u bytes\n", transfer_size, ret) ;

		exit(1) ;
	}

	for(offset = 0 ; offset < transfer_size; offset += SECTOR_SIZE)
	{
		sector = (position + offset) / SECTOR_SIZE ;

		if (DecodeLBA48(&buf[offset])                            != sector)
		{
			(*data_errors)++ ;
			continue ;
		}

		if (verify_code != 0)
		{
			if (DecodeLBA48(&buf[offset + LBA48_SIZE])       != verify_code++)
			{
				(*data_errors)++ ;
				continue ;
			}
		}

		if (DecodeLBA48(&buf[offset + SECTOR_SIZE - LBA48_SIZE]) != sector)
		{
			(*data_errors)++ ;
			continue ;
		}
	}
}

static void Write(int fd, unsigned char *buf, size_t transfer_size, unsigned long long position, unsigned long long verify_code)
{
	unsigned long long	sector ;
	size_t			offset ;
	ssize_t			ret ;

	for(offset = 0 ; offset < transfer_size; offset += SECTOR_SIZE)
	{
		sector = (position + offset) / SECTOR_SIZE ;

		EncodeLBA48(&buf[offset],                            sector) ;
		if (verify_code != 0)
			EncodeLBA48(&buf[offset + LBA48_SIZE],       verify_code++) ;
		EncodeLBA48(&buf[offset + SECTOR_SIZE - LBA48_SIZE], sector) ;
	}

	if ((ret = write(fd, buf, transfer_size)) != transfer_size)
	{
		if (ret < 0)
			perror("write() failed") ;
		else
			fprintf(stderr, "write(%u) returned %u bytes\n", transfer_size, ret) ;

		exit(1) ;
	}
}

void EncodeLBA48(unsigned char *buf, unsigned long long lba)
{
	EncodeValue(buf, LBA48_SIZE, lba) ;
}

unsigned long long DecodeLBA48(unsigned char *buf)
{
	return(DecodeValue(buf, LBA48_SIZE)) ;
}

void EncodeValue(unsigned char *buf, int size, unsigned long long value)
{
	int i ;

	for(i = 0; i < size; i++)
	{
		*(buf++) = value & 0xff ;
		value >>= 8 ;
	}
}

unsigned long long DecodeValue(unsigned char *buf, int size)
{
	unsigned long long	value ;
	int			i ;

	value = 0 ;
	buf += (size - 1) ;

	for(i = 0; i < size; i++)
		value = (value << 8) | *(buf--) ;

	return(value) ;
}
